#!/usr/bin/env python3

'''
Copyright (c) 2020-2023 Children's Hospital of Philadelphia
Author: Li Fang (fangli2718@gmail.com)
              
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

import os
import sys
import time
import argparse
from argparse import RawTextHelpFormatter

from NanoRepeat import nanoRepeat_bam
from NanoRepeat import tk
from NanoRepeat.repeat_region import *
from NanoRepeat import __init__

def map_fastq_to_ref_genome(in_fastq_file, data_type, ref_fasta_file, samtools, minimap2, num_cpu, bam_prefix):
    
    sam_file        = f'{bam_prefix}.sam'
    bam_file        = f'{bam_prefix}.bam'
    sorted_bam_file = f'{bam_prefix}.sorted.bam'

    if os.path.exists(sam_file):
        os.remove(sam_file)

    if os.path.exists(bam_file):
        os.remove(bam_file)

    if os.path.exists(sorted_bam_file):
        os.remove(sorted_bam_file)

    sleep_time = 5

    preset = tk.get_preset_for_minimap2(data_type)
    cmd = f'{minimap2} -a {preset} -t {num_cpu} {ref_fasta_file} {in_fastq_file} > {sam_file} 2> /dev/null'
    tk.run_system_cmd(cmd)
    time.sleep(sleep_time)

    cmd = f'{samtools} view -hb -@ {num_cpu} {sam_file} > {bam_file} 2> /dev/null'
    tk.run_system_cmd(cmd)
    time.sleep(sleep_time)

    if os.path.getsize(bam_file) > 0:
        os.remove(sam_file)
    else:
        tk.eprint(f'ERROR! {bam_file} is empty')
        sys.exit(1)

    cmd = f'{samtools} sort -@ {num_cpu} -o {sorted_bam_file} {bam_file} 2> /dev/null'
    tk.run_system_cmd(cmd)
    time.sleep(sleep_time)

    if os.path.getsize(sorted_bam_file) > 0:
        os.remove(bam_file)
    else:
        tk.eprint(f'ERROR! {sorted_bam_file} is empty')
        sys.exit(1)

    cmd = f'{samtools} index {sorted_bam_file}'
    tk.run_system_cmd(cmd)
    time.sleep(sleep_time)

    return f'{sorted_bam_file}'

def preprocess_fastq(input_args):
    
    bam_prefix    =  f'{input_args.out_prefix}.minimap2'
    in_bam_file   = map_fastq_to_ref_genome(input_args.input, input_args.data_type, input_args.ref_fasta, input_args.samtools, input_args.minimap2, input_args.num_cpu, bam_prefix)
    
    nanoRepeat_bam.nanoRepeat_bam(input_args, in_bam_file)
    return

def main():

    program = 'nanoRepeat.py'
    examples  = f'Examples: \n'
    examples += f'\t# Input is a BAM file:\n'
    examples += f'\t{program} -i input.bam   -t bam   -r hg38.fasta -b hg38.repeats.bed -c 4 -o prefix/of/output/files\n\n'
    examples += f'\t# Input is a FASTQ file:\n'
    examples += f'\t{program} -i input.fastq -t fastq -r hg38.fasta -b hg38.repeats.bed -c 4 -o prefix/of/output/files\n\n'
    examples += f'\t# Input is a FASTA file:\n'
    examples += f'\t{program} -i input.fasta -t fasta -r hg38.fasta -b hg38.repeats.bed -c 4 -o prefix/of/output/files\n\n'
    
    examples += f'\t# For ONT reads basecalled with super accuracy mode:\n'
    examples += f'\t{program} -i input.bam   -t bam -d ont_sup -r hg38.fasta -b hg38.repeats.bed -c 4 -o prefix/of/output/files\n\n'
    examples += f'\t# For ONT Q20+ reads\n'
    examples += f'\t{program} -i input.bam   -t bam -d ont_q20 -r hg38.fasta -b hg38.repeats.bed -c 4 -o prefix/of/output/files\n\n'
    examples += f'\t# For PacBio HiFi reads:\n'
    examples += f'\t{program} -i input.bam   -t bam -d hifi    -r hg38.fasta -b hg38.repeats.bed -c 4 -o prefix/of/output/files\n\n'
    examples += f'\t# For PacBio CLR reads:\n'
    examples += f'\t{program} -i input.bam   -t bam -d clr     -r hg38.fasta -b hg38.repeats.bed -c 4 -o prefix/of/output/files\n\n'
    
    examples += f'\nContact: Li Fang (fangli9@sysu.edu.cn)\n'
    
    parser = argparse.ArgumentParser(prog = program, description=f'NanoRepeat: short tandem repeat (STR) quantification from Nanopore long-read sequencing', epilog=examples, formatter_class=RawTextHelpFormatter)

    # required
    parser.add_argument('-i', '--input', required = True, metavar = 'input_file', type = str, help = '(required) path to input file (supported format: sorted_bam, fastq or fasta)')
    parser.add_argument('-t', '--type', required = True, metavar = 'input_file_type', type = str, help = '(required) input file type (valid values: bam, cram, fastq or fasta)')
    parser.add_argument('-r', '--ref_fasta', required = True, metavar = 'ref.fasta',   type = str, help = '(required) path to reference genome sequence in FASTA format')
    
    parser.add_argument('-b', '--repeat_region_bed', required = True, metavar = 'repeat_regions.bed', type = str, help = '(required) path to the repeat region file (tab delimited text file with 4 columns: chrom start end repeat_unit_seq. Positions start from 0. Start position is self-inclusive but end position is NOT self-inclusive)')
    parser.add_argument('-o', '--out_prefix', required = True, metavar = 'path/to/out_dir/prefix_of_file_names',   type = str, help = '(required) prefix of output files')
    
    # optional
    parser.add_argument('-d', '--data_type', required = False, metavar = 'data_type',  type = str, default = 'ont', help = '(required) sequencing data type. Should be one of the following: ont, ont_sup, ont_q20, clr, hifi')
    parser.add_argument('-c', '--num_cpu', required = False, metavar = 'INT',   type = int, default = 1,  help ='(optional) number of CPU cores (default: 1)')
    parser.add_argument('--samtools', required = False, metavar = 'path/to/samtools',  type = str, default = 'samtools', help ='(optional) path to samtools (default: using environment default)')
    parser.add_argument('--minimap2', required = False, metavar = 'path/to/minimap2',  type = str, default = 'minimap2', help ='(optional) path to minimap2 (default: using environment default)')
    parser.add_argument('--ploidy',   required = False, metavar = 'INT', type = int, default = 2,  help ='(optional) ploidy of the sample (default: 2)')
    parser.add_argument('--anchor_len', required = False, metavar = 'INT', type = int, default = 1000, help ='(optional) length of up/downstream sequence to help identify the repeat region (default: 1000 bp, increase this value if the 1000 bp up/downstream sequences are also repeat)')
    parser.add_argument('--max_mutual_overlap', required = False, metavar = 'FLOAT',  type = float, default = 0.15,  help = 'max mutual overlap of two alleles in terms of repeat size distribution (default value: 0.1). If the Gaussian distribution of two alleles have more overlap than this value, the two alleles will be merged into one allele.')
    parser.add_argument('--remove_noisy_reads', required = False, action='store_true', help = 'remove noisy components when there are more components than ploidy')
    parser.add_argument('--fast_mode', required = False, action='store_true', help = 'fast quantification for repeat size, accuracy may be a little lower')
    parser.add_argument('--save_temp_files', required = False, action='store_true', help = 'save temporary alignment files')
    parser.add_argument('--max_num_components', required = False, metavar = 'INT',  type = int, default = -1,  help = 'max number of components for the Gaussian mixture model (default value: ploidy + 20). Some noisy reads and outlier reads may form a component. Therefore the number of components is usually larger than ploidy. If your sample have too many outlier reads, you can increase this number.')
    
    if len(sys.argv) < 2 or sys.argv[1] in ['help', 'h', '-help', 'usage']:
        input_args = parser.parse_args(['--help'])
    else:
        input_args = parser.parse_args()

    input_args.type = input_args.type.lower()
    
    if input_args.type not in ['bam', 'cram', 'fastq', 'fasta']:
        tk.eprint(f'ERROR! unknown input type: {input_args.type} valid values are: bam, fastq, fasta')
        sys.exit(1)
    
    if input_args.data_type not in ['ont', 'ont_sup', 'ont_q20', 'clr', 'hifi']:
        tk.eprint(f'ERROR! data_type should be one of the following: ont, ont_sup, clr, hifi\n')
        tk.eprint(f'ont:     Oxford Nanopore sequencing, NOT Super Accuracy mode\n')
        tk.eprint(f'ont_sup: Oxford Nanopore sequencing, Super Accuracy mode\n')
        tk.eprint(f'ont_q20: Oxford Nanopore sequencing, Q20+ chemistry\n')
        tk.eprint(f'clr:     PacBio sequencing, Continuous Long Reads (CLR)\n')
        tk.eprint(f'hifi:    PacBio sequencing, HiFi/CCS reads\n')
        sys.exit(1)
        
    if input_args.ploidy < 1:
        tk.eprint(f'ERROR! ploidy should be at least 1')
        sys.exit(1)

    if input_args.max_mutual_overlap >= 1.0:
        tk.eprint('ERROR! --max_mutual_overlap must be < 1\n')
        sys.exit(1)

    if input_args.max_num_components == -1:
        input_args.max_num_components = input_args.ploidy + 20

    out_dir, out_file_prefix = os.path.split(input_args.out_prefix)
    out_dir = os.path.abspath(out_dir)
    if out_file_prefix == '':
        out_file_prefix = os.path.split(input_args.input)[1]
        input_args.out_prefix = os.path.join(out_dir, out_file_prefix)
    else:
        input_args.out_prefix = os.path.abspath(input_args.out_prefix)

    input_args.input             = os.path.abspath(input_args.input)
    input_args.ref_fasta         = os.path.abspath(input_args.ref_fasta)
    input_args.out_prefix        = os.path.abspath(input_args.out_prefix)
    input_args.repeat_region_bed = os.path.abspath(input_args.repeat_region_bed)

    tk.eprint(f'NOTICE: Input file is: {input_args.input}')
    tk.eprint(f'NOTICE: Input type is: {input_args.type}')
    tk.eprint(f'NOTICE: Referece fasta file is: {input_args.ref_fasta}')
    tk.eprint(f'NOTICE: Output prefix is: {input_args.out_prefix}')
    tk.eprint(f'NOTICE: Repeat region bed file is: {input_args.repeat_region_bed}')

    os.makedirs(out_dir, exist_ok=True)

    if input_args.type == 'fastq':
        preprocess_fastq(input_args)
    elif input_args.type == 'fasta':
        preprocess_fastq(input_args)
    elif input_args.type == 'bam' or input_args.type == 'cram':
        nanoRepeat_bam.nanoRepeat_bam(input_args, input_args.input)
    else:
        tk.eprint(f'ERROR! Unknown input type: {input_args.type} valid values are: bam, cram, fastq, fasta')
        sys.exit(1)
    
    return


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

'''
Copyright (c) 2020-2023 Children's Hospital of Philadelphia
Author: Li Fang (fangli2718@gmail.com)
              
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''

import os
import string
import sys
import shutil
import numpy as np
from multiprocessing import Process
from typing import List

from NanoRepeat import tk
from NanoRepeat.repeat_region import *
from NanoRepeat.paf import *
from NanoRepeat.split_alleles import *

class Metainfo:
    def __init__(self, allele_id = -1, num_reads = 0, predicted_repeat_size = -1, min_repeat_size = -1, max_repeat_size = -1):
        self.allele_id = allele_id
        self.num_reads = num_reads
        self.predicted_repeat_size = predicted_repeat_size
        self.min_repeat_size = min_repeat_size
        self.max_repeat_size = max_repeat_size
    
    def init_from_repeat_size_list(self, repeat_size_list, allele_id):
        
        self.allele_id = allele_id
        self.num_reads = len(repeat_size_list)
        if len(repeat_size_list) == 0:
            self.predicted_repeat_size = -1
            self.min_repeat_size = -1 
            self.max_repeat_size = -1
            return self
        
        self.min_repeat_size = min(repeat_size_list)
        self.max_repeat_size = max(repeat_size_list)
        self.predicted_repeat_size = int(np.median(repeat_size_list) + 0.5)
        return self

    def output(self):
        return 'allele_id=%d\tnum_reads=%d\tpredicted_repeat_size=%d\tmin_repeat_size=%d\tmax_repeat_size=%d' % (self.allele_id, self.num_reads, self.predicted_repeat_size, self.min_repeat_size, self.max_repeat_size)


def extract_ref_sequence(ref_fasta_dict, repeat_region:RepeatRegion):

    MIN_ANCHOR_LEN = 10
    if repeat_region.anchor_len < MIN_ANCHOR_LEN:
        tk.eprint('WARNING: anchor_len must be >= 10!')
        repeat_region.anchor_len = 10

    chr_name = repeat_region.chrom
    if chr_name not in ref_fasta_dict:
        if 'chr' == chr_name[0:3]:
            chr_name = chr_name[3:]
        else:
            chr_name = 'chr' + chr_name
    if chr_name not in ref_fasta_dict:
        tk.eprint(f'ERROR! the following chromosome in repeat region bed file was not found in the reference fasta file: {repeat_region.chrom}')
        sys.exit()
    
    chr_seq = ref_fasta_dict[chr_name]
    chr_len = len(chr_seq)

    if repeat_region.start_pos > chr_len:
        tk.eprint(f'ERROR! the repeat start position is larger than chromosome length: {repeat_region.start_pos}. chromosome length = {chr_len}')
        sys.exit()
    
    if repeat_region.start_pos < 0:
        tk.eprint(f'ERROR! the repeat start position < 0')
        sys.exit()

    if repeat_region.end_pos > chr_len + 1:
        tk.eprint(f'ERROR! the repeat end position is larger than chromosome length: {repeat_region.end_pos}. chromosome length = {chr_len}')
        sys.exit()
    
    if repeat_region.end_pos < repeat_region.start_pos:
        tk.eprint(f'ERROR! end position is smaller than start position: {repeat_region.to_invertal()}')
        sys.exit()

    start_pos = repeat_region.start_pos - repeat_region.anchor_len
    end_pos = repeat_region.end_pos + repeat_region.anchor_len
    if start_pos < 0: 
        start_pos = 0
    
    if end_pos > chr_len:
        end_pos = chr_len
    
    repeat_region.left_anchor_seq  = chr_seq[start_pos:repeat_region.start_pos]
    repeat_region.left_anchor_len  = len(repeat_region.left_anchor_seq)

    repeat_region.right_anchor_seq = chr_seq[repeat_region.end_pos:end_pos]
    repeat_region.right_anchor_len = len(repeat_region.right_anchor_seq)

    repeat_region.mid_ref_seq      = chr_seq[repeat_region.start_pos:repeat_region.end_pos] # repeat sequence in the ref genome

    if repeat_region.left_anchor_len == 0 and repeat_region.right_anchor_len == 0:
        tk.eprint('ERROR! there is no flanking sequence around the repeat region! ')
        sys.exit(1)

    if repeat_region.left_anchor_len < MIN_ANCHOR_LEN and repeat_region.right_anchor_len < MIN_ANCHOR_LEN:
        tk.eprint(f'ERROR! both left and right flanking sequences are less than {MIN_ANCHOR_LEN} bp. The quantification might not be accurate!')
        sys.exit(1)
    
    return

def refine_repeat_region_in_ref(minimap2:string, repeat_region:RepeatRegion):

    mid_ref_seq            = repeat_region.mid_ref_seq
    repeat_unit_seq        = repeat_region.repeat_unit_seq
    repeat_region_location = repeat_region.to_invertal()
    temp_out_dir           = repeat_region.temp_out_dir
    out_prefix             = repeat_region.out_prefix
    error_message_file     = out_prefix + '.error_message.log'
    mid_ref_seq_len        = len(mid_ref_seq)

    buffer_size = 10
    if mid_ref_seq_len < buffer_size: return

    num_units = int(float(mid_ref_seq_len) / len(repeat_unit_seq))+1
    pure_repeat_seq = repeat_unit_seq * num_units

    ref_check_paf_file   = os.path.join(temp_out_dir, 'ref_aligned_to_pure_repeats.paf')
    mid_ref_seq_file     = os.path.join(temp_out_dir, 'mid_ref_seq.fasta')
    pure_repeat_seq_file = os.path.join(temp_out_dir, 'pure_repeat_seq.fasta')

    write_seq_to_fasta(f'mid_ref_seq_{repeat_region_location}', mid_ref_seq, mid_ref_seq_file)
    write_seq_to_fasta('pure_repeat_seq', pure_repeat_seq, pure_repeat_seq_file)

    score_parameters = '-x ava-ont -z30 -k3 -w2 -m1 -n2 -s10'

    cmd = f'{minimap2} {score_parameters} -f 0 --cs  {pure_repeat_seq_file} {mid_ref_seq_file} > {ref_check_paf_file}'
    tk.run_system_cmd(cmd)

    ref_check_paf_f = open(ref_check_paf_file, 'r')
    lines = list(ref_check_paf_f)
    ref_check_paf_f.close()

    paf_list = []
    for line in lines:
        col_list = line.strip().split('\t')
        paf = PAF(col_list)
        paf_list.append(paf)
    
    if len(paf_list) == 0:
        repeat_region.ref_has_issue = True
        
        error_message = f'ERROR! The given repeat motif ({repeat_region.repeat_unit_seq}) was not found in the reference sequence ({repeat_region_location}). Please check your input!\n'
        error_message += f'The repeat motif is: {repeat_unit_seq}\n'
        error_message += f'The sequence inside {repeat_region_location} is: {mid_ref_seq}\n'
        
        tk.eprint(f'current repeat: {repeat_region.to_unique_id()}\n')
        tk.eprint(error_message)
        error_message_f = open(error_message_file, 'w')
        error_message_f.write(error_message)
        error_message_f.close()

        return

    max_align_score     = paf_list[0].align_score
    max_align_score_idx = 0
    for i in range(1, len(paf_list)):
        if paf_list[i].align_score > max_align_score:
            max_align_score = paf_list[i].align_score
            max_align_score_idx = i
    
    best_paf  = paf_list[max_align_score_idx]
    align_len = best_paf.qend - best_paf.qstart

    if best_paf.align_score < 0:
        repeat_region.ref_has_issue = True
        error_message  = f'ERROR! In the reference sequence, the interval ({repeat_region_location}) has too many mismatches with the repeat motif. Please check your input.\n\n'
        error_message += f'The repeat motif is: {repeat_unit_seq}\n'
        error_message += f'The sequence inside {repeat_region_location} is:\n'
        error_message += mid_ref_seq[0:best_paf.qstart] + '---' + mid_ref_seq[best_paf.qstart:best_paf.qend] + '---' + mid_ref_seq[best_paf.qend:] + '\n'
        
        tk.eprint(f'current repeat: {repeat_region.to_unique_id()}\n')
        tk.eprint(error_message)
        error_message_f = open(error_message_file, 'w')
        error_message_f.write(error_message)
        error_message_f.close()
        return
        
    if mid_ref_seq_len - align_len > buffer_size and float(align_len)/mid_ref_seq_len < 0.8:
        repeat_region.ref_has_issue = True
        
        corrected_start_pos = repeat_region.start_pos + best_paf.qstart
        corrected_end_pos   = corrected_start_pos + align_len
        
        error_message = f'ERROR! In the reference sequence, the interval ({repeat_region_location}) included non-repeat sequences or repeats that are not the given motif. Please check your input.\n'
        error_message += f'The repeat motif is: {repeat_unit_seq}\n'
        error_message += f'The sequence inside {repeat_region_location} is:\n'
        error_message += mid_ref_seq[0:best_paf.qstart] + '---' + mid_ref_seq[best_paf.qstart:best_paf.qend] + '---' + mid_ref_seq[best_paf.qend:] + '\n'
        error_message += f'\n\nWe suggest you change the chromosome positions (in the `--repeat_region_bed` file) to:\n{repeat_region.chrom}\t{corrected_start_pos}\t{corrected_end_pos}\t{repeat_unit_seq}\n'
        
        tk.eprint(f'current repeat: {repeat_region.to_unique_id()}\n')
        tk.eprint(error_message)
        error_message_f = open(error_message_file, 'w')
        error_message_f.write(error_message)
        error_message_f.close()

    return

def write_seq_to_fasta(seq_name, seq, out_file):

    out_f = open(out_file, 'w')
    out_f.write(f'>{seq_name}\n')
    out_f.write(f'{seq}\n')
    out_f.close()

    return

def check_anchor_mapping(one_read_anchor_paf_list: List[PAF]):

    if len(one_read_anchor_paf_list) == 0:
        return False
    
    if len(one_read_anchor_paf_list) == 1:
        return True
    
    if one_read_anchor_paf_list[0].align_len < 10: 
        return False
    
    if one_read_anchor_paf_list[0].align_score > 1.5 * one_read_anchor_paf_list[1].align_score and one_read_anchor_paf_list[0].mapq > 30:
        return True
    else:
        return False

def find_anchor_locations_for1read(read_paf_list: List[PAF], repeat_region: RepeatRegion):

    if len(read_paf_list) == 0: return

    left_anchor_paf_list  = []
    right_anchor_paf_list = []
    for paf in read_paf_list:
        if paf.tname.startswith('left_anchor'):
            left_anchor_paf_list.append(paf)
        elif paf.tname.startswith('right_anchor'):
            right_anchor_paf_list.append(paf)

    left_anchor_paf_list.sort(key = lambda paf:paf.align_score, reverse=True)
    right_anchor_paf_list.sort(key = lambda paf:paf.align_score, reverse=True)
    read = Read()
    read.read_name = read_paf_list[0].qname
    read.full_read_len = read_paf_list[0].qlen
    read.both_anchors_are_good = False
    read.left_anchor_is_good = check_anchor_mapping(left_anchor_paf_list)
    read.right_anchor_is_good = check_anchor_mapping(right_anchor_paf_list)
    
    if read.left_anchor_is_good == False: return
    if read.right_anchor_is_good == False: return
    
    repeat_region_length = 0
    read.left_anchor_paf = left_anchor_paf_list[0]
    read.right_anchor_paf = right_anchor_paf_list[0]

    # Important: if strand == '-', all positions are based on the reverse strand
    if read.left_anchor_is_good and read.right_anchor_is_good and read.left_anchor_paf.strand == read.right_anchor_paf.strand:
        repeat_region_length = read.right_anchor_paf.qstart - read.left_anchor_paf.qend

    if repeat_region_length > -10:
        read.both_anchors_are_good = True
        read.dist_between_anchors = repeat_region_length
    
    if read.both_anchors_are_good == False: return

    repeat_region.read_dict[read.read_name] = read

    repeat_region.buffer_len = 100
    if read.both_anchors_are_good:
        read.core_seq_start_pos = read.left_anchor_paf.qend - repeat_region.buffer_len
        read.core_seq_end_pos   = read.right_anchor_paf.qstart + repeat_region.buffer_len
        read.mid_seq_start_pos  = read.left_anchor_paf.qend
        read.mid_seq_end_pos    = read.right_anchor_paf.qstart
        if read.core_seq_start_pos < 0: read.core_seq_start_pos = 0
        if read.core_seq_end_pos > read.full_read_len: read.core_seq_end_pos = read.full_read_len
        read.left_buffer_len = read.left_anchor_paf.qend - read.core_seq_start_pos
        read.right_buffer_len = read.core_seq_end_pos - read.right_anchor_paf.qstart
        if read.left_anchor_paf.strand == '+':
            read.strand = '+'
        else:
            read.strand = '-'

    return

def find_anchor_locations_from_paf(repeat_region: RepeatRegion, anchor_locations_paf_file: string):

    anchor_locations_paf_f = open(anchor_locations_paf_file, 'r')
    read_paf_list = list()
    while 1:
        line = anchor_locations_paf_f.readline()
        if not line: break
        line = line.strip()
        if not line: continue

        col_list = line.strip().split('\t')
        paf = PAF(col_list)

        if len(read_paf_list) == 0 or paf.qname == read_paf_list[0].qname:
            read_paf_list.append(paf)
        else:
            find_anchor_locations_for1read(read_paf_list, repeat_region)
            read_paf_list.clear()
            read_paf_list.append(paf)

    anchor_locations_paf_f.close()
    
    find_anchor_locations_for1read(read_paf_list, repeat_region)

    return

def find_anchor_locations_in_reads(minimap2:string, data_type:string, repeat_region:RepeatRegion, num_cpu:int):

    left_template_seq    = repeat_region.left_anchor_seq
    left_template_name   = 'left_anchor'

    right_template_seq   = repeat_region.right_anchor_seq
    right_template_name  = 'right_anchor'

    template_fasta_file  = os.path.join(repeat_region.temp_out_dir, 'anchors.fasta')
    repeat_region.temp_file_list.append(template_fasta_file)

    template_fasta_fp    = open(template_fasta_file, 'w')
    template_fasta_fp.write('>%s\n' % left_template_name)
    template_fasta_fp.write('%s\n' % left_template_seq)
    template_fasta_fp.write('>%s\n' % right_template_name)
    template_fasta_fp.write('%s\n' % right_template_seq)
    template_fasta_fp.close()

    anchor_locations_paf_file = os.path.join(repeat_region.temp_out_dir, 'anchor_locations.paf')
    repeat_region.temp_file_list.append(anchor_locations_paf_file)

    preset = tk.get_preset_for_minimap2(data_type)
    cmd = f'{minimap2} -c -t {num_cpu} {preset} {template_fasta_file} {repeat_region.region_fq_file} > {anchor_locations_paf_file}'
    tk.run_system_cmd(cmd)

    find_anchor_locations_from_paf(repeat_region, anchor_locations_paf_file)
    
    return

def make_core_seq_fastq(repeat_region: RepeatRegion):
    repeat_region.core_seq_fq_file = os.path.join(repeat_region.temp_out_dir, 'core_sequences.fastq')
    repeat_region.temp_file_list.append(repeat_region.core_seq_fq_file)

    repeat_region.mid_seq_fq_file = os.path.join(repeat_region.temp_out_dir, 'middle_sequences.fastq')
    repeat_region.temp_file_list.append(repeat_region.mid_seq_fq_file)

    region_fq_f   = open(repeat_region.region_fq_file, 'r')
    core_seq_fq_f = open(repeat_region.core_seq_fq_file, 'w')
    mid_seq_fq_f  = open(repeat_region.mid_seq_fq_file, 'w')
    while 1:
        line1 = region_fq_f.readline()
        line2 = region_fq_f.readline()
        line3 = region_fq_f.readline()
        line4 = region_fq_f.readline()
        if not line1: break
        if not line2: break
        if not line3: break
        if not line4: break

        read_name = line1.strip()[1:]
        if read_name not in repeat_region.read_dict: continue
        read_sequence = line2.strip()            
        if repeat_region.read_dict[read_name].strand == '-':
            read_sequence = tk.rev_comp(read_sequence)
        core_sequence = read_sequence[repeat_region.read_dict[read_name].core_seq_start_pos:repeat_region.read_dict[read_name].core_seq_end_pos]
        mid_sequence = read_sequence[repeat_region.read_dict[read_name].mid_seq_start_pos:repeat_region.read_dict[read_name].mid_seq_end_pos]

        repeat_region.read_core_seq_dict[read_name] = core_sequence
        core_seq_fq_f.write(line1)
        core_seq_fq_f.write(core_sequence + '\n')
        core_seq_fq_f.write(line3)
        core_seq_fq_f.write('0' * len(core_sequence) + '\n')


        mid_seq_fq_f.write(line1)
        mid_seq_fq_f.write(mid_sequence + '\n')
        mid_seq_fq_f.write(line3)
        mid_seq_fq_f.write('0' * len(mid_sequence) + '\n')

    region_fq_f.close()
    core_seq_fq_f.close()
    mid_seq_fq_f.close()
    return


def round1_and_round2_estimation(minimap2:string, data_type:string, repeat_region: RepeatRegion, num_cpu: int):

    if len(repeat_region.read_dict) == 0: return

    round1_repeat_size_list = []
    for read_name in repeat_region.read_dict:
        read = repeat_region.read_dict[read_name]
        read.round1_repeat_size = float(read.dist_between_anchors)/len(repeat_region.repeat_unit_seq)
        round1_repeat_size_list.append(read.round1_repeat_size)
    
    template_repeat_size = int(max(round1_repeat_size_list) * 1.5) + 1

    if template_repeat_size < max(round1_repeat_size_list) + 10:
        template_repeat_size = int(max(round1_repeat_size_list) + 10)
    
    round1_fasta_file  = os.path.join(repeat_region.temp_out_dir, 'round1_ref.fasta')
    repeat_region.temp_file_list.append(round1_fasta_file)
    
    round1_fasta_f = open(round1_fasta_file, 'w')
    round1_fasta_f.write(f'>{template_repeat_size}\n')
    round1_fasta_f.write(f'{repeat_region.left_anchor_seq}{repeat_region.repeat_unit_seq * template_repeat_size}\n')
    round1_fasta_f.close()

    round1_paf_file = os.path.join(repeat_region.temp_out_dir, 'round1.paf')
    repeat_region.temp_file_list.append(round1_paf_file)
    
    preset = tk.get_preset_for_minimap2(data_type)
    cmd = f'{minimap2} -c -t {num_cpu} {preset} -f 0.0 {round1_fasta_file} {repeat_region.core_seq_fq_file} > {round1_paf_file}'
    tk.run_system_cmd(cmd)
    round1_paf_f = open(round1_paf_file, 'r')
    lines = list(round1_paf_f)
    round1_paf_f.close()


    round2_repeat_size_dict = dict()
    
    for line in lines:
        col_list = line.strip().split('\t')
        paf = PAF(col_list)

        if paf.tstart <= len(repeat_region.left_anchor_seq) and paf.tend >= len(repeat_region.left_anchor_seq):
            read_name = paf.qname
            round2_repeat_size = float(paf.tend - len(repeat_region.left_anchor_seq))/len(repeat_region.repeat_unit_seq)
            align_score = paf.align_score
            if read_name not in round2_repeat_size_dict:
                round2_repeat_size_dict[read_name] = []
            round2_repeat_size_dict[read_name].append((align_score, round2_repeat_size))
            
    for read_name in round2_repeat_size_dict:
        round2_repeat_size_dict[read_name].sort(key = lambda x:x[0], reverse=True)
        round2_repeat_size = round2_repeat_size_dict[read_name][0][1]
        repeat_region.read_dict[read_name].round2_repeat_size = round2_repeat_size

    round2_repeat_size_file = os.path.join(repeat_region.temp_out_dir, 'round2_repeat_size.txt')
    round2_repeat_size_f = open(round2_repeat_size_file, 'w')
    for read_name in repeat_region.read_dict:
        round2_repeat_size_f.write(f'{read_name}\t{repeat_region.read_dict[read_name].round2_repeat_size}\n')
    
    round2_repeat_size_f.close()
    
    return

def get_max_read_length_from_fastq(fq_file):

    max_read_length = 0
    fq_f = open(fq_file, 'r')
    while 1:
        line = fq_f.readline()
        if not line: break
        if len(line)-1 > max_read_length:
            max_read_length = len(line)-1
    fq_f.close()

    return max_read_length

def round3_estimation_for1read(repeat_region: RepeatRegion, read_paf_list):

    if len(read_paf_list) == 0: return

    read_paf_list.sort(key = lambda paf:paf.align_score, reverse = True)
    best_paf_repeat_size_list = []
    read_name = read_paf_list[0].qname
    for i in range(0, len(read_paf_list)):
        if read_paf_list[i].align_score < read_paf_list[0].align_score: break
        if read_paf_list[i].tstart < repeat_region.left_anchor_len and read_paf_list[i].tlen - read_paf_list[i].tend < repeat_region.right_anchor_len and read_paf_list[i].align_score == read_paf_list[0].align_score:
            best_paf_repeat_size_list.append(int(read_paf_list[i].tname))

    if len(best_paf_repeat_size_list) > 0:
        repeat_region.read_dict[read_name].round3_repeat_size = np.mean(best_paf_repeat_size_list)
    else:
        repeat_region.read_dict[read_name].round3_repeat_size = repeat_region.read_dict[read_name].round2_repeat_size
    return
                
def round3_estimation_from_paf(repeat_region: RepeatRegion, round3_paf_file):

    round3_paf_f = open(round3_paf_file, 'r')
    lines = list(round3_paf_f)
    round3_paf_f.close()

    read_paf_list = []
    for line in lines:
        col_list = line.strip().split('\t')
        paf = PAF(col_list)
        if paf.qname not in repeat_region.read_dict: continue
        if repeat_region.read_dict[paf.qname].round2_repeat_size == None: continue
        if len(read_paf_list) == 0 or paf.qname == read_paf_list[0].qname:
            read_paf_list.append(paf)
        else:
            round3_estimation_for1read(repeat_region, read_paf_list)
            read_paf_list.clear()
            read_paf_list.append(paf)

    round3_estimation_for1read(repeat_region, read_paf_list)
    
    return 

def round3_estimation(minimap2:string, data_type:string, fast_mode, repeat_region:RepeatRegion, num_cpu:int):
    
    round3_paf_file = os.path.join(repeat_region.temp_out_dir, 'round3.paf')
    round3_paf_f = open(round3_paf_file, 'w')
    round3_paf_f.close()
    
    procs = []

    for proc_id in range(0, num_cpu):
        proc = Process(target=round3_align_1process, args=(proc_id, num_cpu, fast_mode, repeat_region, data_type, minimap2))
        procs.append(proc)
        proc.start()
        
    for proc_id in range(0, num_cpu):
        procs[proc_id].join()
        proc_out_paf_file = os.path.join(repeat_region.temp_out_dir, f'round3_align.process{proc_id}.paf')
        cmd = f'cat {proc_out_paf_file} >> {round3_paf_file}'
        tk.run_system_cmd(cmd)
        os.remove(proc_out_paf_file)
    
    round3_estimation_from_paf(repeat_region, round3_paf_file)

def round3_align_1process(proc_id:int, num_cpu:int, fast_mode, repeat_region:RepeatRegion, data_type:string, minimap2:string):
    
    read_id = -1
    out_paf_file = os.path.join(repeat_region.temp_out_dir, f'round3_align.process{proc_id}.paf')
    out_paf_f = open(out_paf_file, 'w')
    out_paf_f.close()
    
    for read_name in repeat_region.read_dict:
        read = repeat_region.read_dict[read_name]
        round2_repeat_size = read.round2_repeat_size
        if round2_repeat_size == None: continue
        read_id += 1
        if read_id % num_cpu != proc_id: continue
        
        buffer = max(15, int(round2_repeat_size * 0.05))
        if buffer > 150: buffer = 150
        
        if fast_mode:
            buffer = 15
        
        max_template_repeat_size = int(round2_repeat_size + buffer)
        min_template_repeat_size = int(round2_repeat_size - buffer)
        
        if min_template_repeat_size < 0: min_template_repeat_size = 0
        
        template_fasta_file  = os.path.join(repeat_region.temp_out_dir, f'round3_reference.read{read_id}.fasta')
        template_fasta_fp    = open(template_fasta_file, 'w')
        for repeat_size in range(min_template_repeat_size, max_template_repeat_size+1):
            template_seq = repeat_region.left_anchor_seq + repeat_region.repeat_unit_seq * repeat_size + repeat_region.right_anchor_seq
            template_fasta_fp.write('>%s\n' % repeat_size)
            template_fasta_fp.write('%s\n' % template_seq)

        template_fasta_fp.close()
        
        read_seq = repeat_region.read_core_seq_dict[read_name].strip()
        read_fasta_file = os.path.join(repeat_region.temp_out_dir, f'round3_input.read{read_id}.fasta')
        
        read_fasta_f = open(read_fasta_file, 'w')
        read_fasta_f.write(f'>{read_name}\n')
        read_fasta_f.write(read_seq + '\n')
        read_fasta_f.close()

        preset = tk.get_preset_for_minimap2(data_type)
        cmd = f'{minimap2} {preset} -f 0.0 -N 100 -c --eqx -t 1 {template_fasta_file} {read_fasta_file} >> {out_paf_file}'
        tk.run_system_cmd(cmd)

        os.remove(template_fasta_file)
        os.remove(read_fasta_file)

    return

def remove_noisy_reads_1d(allele_list, ploidy):

    tk.eprint('NOTICE: Trying to remove noisy reads')
    allele_list.sort(key = lambda allele:allele.num_reads)
    num_removed_reads = 0
    while len(allele_list) > ploidy and len(allele_list) >= 2:
        if allele_list[0].num_reads * 1.5 <= allele_list[-ploidy].num_reads:
            num_removed_reads += allele_list[0].num_reads 
            allele_list.pop(0)
        else:
            break

    return allele_list, num_removed_reads


def split_allele_using_gmm_1d(repeat_region, ploidy, error_rate, max_mutual_overlap, max_num_components, remove_noisy_reads):

    in_fastq_file = repeat_region.region_fq_file
    out_prefix = repeat_region.out_prefix

    if ploidy < 1:
        tk.eprint('ERROR! ploidy must be >= 1 !')
        sys.exit()


    read_repeat_count_dict = dict()
    for read_name in repeat_region.read_dict:
        if repeat_region.read_dict[read_name].round3_repeat_size != None:
            read_repeat_count_dict[read_name] = repeat_region.read_dict[read_name].round3_repeat_size

    if len(read_repeat_count_dict) == 0:
        tk.eprint(f'ERROR! No reads were found for repeat region: {repeat_region.to_outfile_prefix()}')
        return

    if len(read_repeat_count_dict) == 1:
        tk.eprint(f'ERROR! No enough reads for phasing. Repeat region is: {repeat_region.to_outfile_prefix()}')
        return

    probability_cutoff = 0.95
    cov_type = 'diag'
    dimension = 1
    
    readname_list, read_repeat_count_list = remove_outlier_reads_1d(read_repeat_count_dict)
    read_repeat_count_array = np.array(read_repeat_count_list).reshape(-1, 1)

    simulated_read_repeat_count_list = simulate_reads(read_repeat_count_list, error_rate)
    simualted_read_repeat_count_array = np.array(simulated_read_repeat_count_list).reshape(-1, dimension)

    best_n_components, final_gmm = auto_GMM_1d (simualted_read_repeat_count_array, max_num_components, cov_type, max_mutual_overlap)
    tk.eprint(f'NOTCIE: Number of alleles={best_n_components}')
    
    allele_list = create_allele_list_1d(best_n_components, final_gmm, readname_list, read_repeat_count_array, read_repeat_count_dict, probability_cutoff)
    
    num_removed_reads = 0
    if remove_noisy_reads == True and len(allele_list) > ploidy:
        allele_list, num_removed_reads = remove_noisy_reads_1d(allele_list, ploidy)
        tk.eprint(f'NOTICE: There are {len(allele_list)} alleles after removing noisy reads')
    
    best_n_components = len(allele_list)
    allele_list.sort(key = lambda allele:allele.gmm_mean1)

    readinfo_dict = create_readinfo_dict_from_allele_list(allele_list, dimension)

    tk.eprint('NOTICE: Writing phasing results...')
    output_phasing_results_1d(allele_list, repeat_region.to_unique_id(), repeat_region.out_prefix)
    
    output_phased_fastq(in_fastq_file, readinfo_dict, best_n_components, out_prefix)

    tk.eprint('NOTICE: Writing summary file...')
    output_summary_file_1d(allele_list, repeat_region.to_unique_id(), num_removed_reads, out_prefix)

    tk.eprint('NOTICE: Plotting figures...')
    plot_repeat_counts_1d(readinfo_dict, allele_list, repeat_region.to_unique_id(), out_prefix)

    return
    
def quantify1repeat_from_bam(input_args, error_rate, in_bam_file:string, ref_fasta_dict:dict, repeat_region:RepeatRegion):

    temp_out_dir = f'{input_args.out_prefix}.NanoRepeat_temp_dir.{repeat_region.to_outfile_prefix()}'
    os.makedirs(temp_out_dir, exist_ok=True)

    repeat_region.temp_out_dir = temp_out_dir
    repeat_region.out_prefix = f'{input_args.out_prefix}.{repeat_region.to_outfile_prefix()}'
    repeat_region.anchor_len = input_args.anchor_len

    # extract reads from bam file
    repeat_region.region_fq_file = os.path.join(temp_out_dir, f'{repeat_region.to_outfile_prefix()}.fastq')
    region_bam_file = f'{repeat_region.out_prefix}.sorted.bam'
    cmd = f'{input_args.samtools} view --reference {input_args.ref_fasta} -hb {in_bam_file} {repeat_region.to_invertal(flank_dist=repeat_region.anchor_len)} > {region_bam_file}'
    tk.run_system_cmd(cmd)
    
    cmd = f'{input_args.samtools} index {region_bam_file}'
    tk.run_system_cmd(cmd)
    
    cmd = f'{input_args.samtools} fastq {region_bam_file} > {repeat_region.region_fq_file}'
    tk.run_system_cmd(cmd)

    fastq_file_size = os.path.getsize(repeat_region.region_fq_file)
    if fastq_file_size == 0:
        tk.eprint(f'WARNING! No reads were found in repeat region: {repeat_region.to_outfile_prefix()}')
        if input_args.save_temp_files == False: 
            shutil.rmtree(repeat_region.temp_out_dir)
        return

    # extract ref sequence
    extract_ref_sequence(ref_fasta_dict, repeat_region)
    
    refine_repeat_region_in_ref(input_args.minimap2, repeat_region)
    
    if repeat_region.ref_has_issue == True: 
        if input_args.save_temp_files == False: 
            shutil.rmtree(repeat_region.temp_out_dir)
        return
    
    tk.eprint('NOTICE: Step 1: finding anchor location in reads')
    find_anchor_locations_in_reads(input_args.minimap2, input_args.data_type, repeat_region, input_args.num_cpu)
    tk.eprint('NOTICE: Step 1 finished')

    # make core sequence fastq
    make_core_seq_fastq(repeat_region)

    tk.eprint('NOTICE: Step 2: round 1 and round 2 estimation')
    round1_and_round2_estimation(input_args.minimap2, input_args.data_type, repeat_region, input_args.num_cpu)
    tk.eprint('NOTICE: Step 2 finished')

    tk.eprint('NOTICE: Step 3: round 3 estimation')
    round3_estimation(input_args.minimap2, input_args.data_type, input_args.fast_mode, repeat_region, input_args.num_cpu)
    tk.eprint('NOTICE: Step 3 finished')

    tk.eprint('NOTICE: Writing to repeat size file...')
    output_repeat_size_1d(repeat_region)

    tk.eprint('NOTICE: Step 4: phasing reads using GMM')
    split_allele_using_gmm_1d(repeat_region, input_args.ploidy, error_rate, input_args.max_mutual_overlap, input_args.max_num_components, input_args.remove_noisy_reads)
    
    if input_args.save_temp_files == False:
        shutil.rmtree(repeat_region.temp_out_dir)

    return

def nanoRepeat_bam (input_args, in_bam_file:string):
    
    # ont, ont_sup, ont_q20, clr, hifi

    if input_args.data_type == 'ont' or 'clr':
        error_rate = 0.07
    elif input_args.data_type == 'ont_sup':
        error_rate = 0.04
    elif input_args.data_type == 'ont_q20':
        error_rate = 0.03
    elif input_args.data_type == 'hifi':
        error_rate = 0.02
    else:
        tk.eprint(f'ERROR! unknown data type: {input_args.data_type}')
        sys.exit(1)
    
    tk.eprint(f'NOTICE: Reading repeat region file: {input_args.repeat_region_bed}')
    repeat_region_list = read_repeat_region_file(input_args.repeat_region_bed)

    tk.eprint(f'NOTICE: Reading reference fasta file: {input_args.ref_fasta}')
    ref_fasta_dict = tk.fasta_file2dict(input_args.ref_fasta)
    
    for repeat_region in repeat_region_list:
        tk.eprint(f'NOTICE: Quantifying repeat: {repeat_region.to_outfile_prefix()}')
        quantify1repeat_from_bam(input_args, error_rate, in_bam_file, ref_fasta_dict, repeat_region)

    tk.eprint('NOTICE: Program finished.')

    return




#!/usr/bin/env python3


'''
Copyright (c) 2020-2023 Children's Hospital of Philadelphia
Author: Li Fang (fangli2718@gmail.com)
              
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''


import os
import sys
from typing import final
import numpy as np

import argparse
import shutil


from NanoRepeat import tk
from NanoRepeat.paf import *
from NanoRepeat.split_alleles import *

class Repeat:
    def __init__(self):
        self.repeat_id = ''
        self.chrom = ''
        self.start = -1
        self.end = -1
        self.repeat_unit = ''
        self.repeat_unit_size = 0
        self.min_size = 0
        self.max_size = 1000
        self.round1_min_size = 0
        self.round1_max_size = 1000
        self.is_main = 0
    
    def init_from_string(self, string):
        col_list = string.split(':')
        if len(col_list) != 5:
            tk.eprint('ERROR! --repeat1 and --repeat2 should be in this format: chr:start:end:repeat_unit:max_size (coordinates are 0-based, e.g. chr4:3074876:3074933:CAG:200).')
            sys.exit(1)
        
        self.chrom, self.start, self.end, self.repeat_unit, self.max_size = col_list
        self.start = int(self.start)
        self.end   = int(self.end)
        self.repeat_unit_size = len(self.repeat_unit)
        self.min_size = 0
        self.max_size = int(self.max_size)
        self.repeat_id = '-'.join(col_list[0:4])
        return self

class Round1Estimation:
    def __init__(self):
        self.repeat1_count_range_dict = dict()
        self.repeat2_count_range_dict = dict()
        self.potential_repeat_region_dict = dict()
        self.bad_reads_set = set()
    
    def output_repeat2_boundaries(self):
        out_string = ''
        for readname in self.repeat2_count_range_dict:
            lower_bound, upper_bound = self.repeat2_count_range_dict[readname]
            out_string += f'{readname}\t{lower_bound}\t{upper_bound}\n'

        return out_string

class RepeatSize:
    def __init__(self):
        self.repeat1_count_dict = dict()
        self.repeat2_count_dict = dict()
        self.step_size1 = 1
        self.step_size2 = 1

def parse_user_arguments():

    parser = argparse.ArgumentParser(description='Joint quantification of two adjacent tandem repeats from long-read amplicon sequencing data ')
    ### required arguments ###
    parser.add_argument('-i', '--in_fq',        required = True,  metavar = 'PATH',   type = str, help = 'input fastq file')
    parser.add_argument('-r', '--ref_fasta',    required = True,  metavar = 'PATH',   type = str, help = 'reference genome sequence in FASTA format')
    parser.add_argument('-1', '--repeat1',      required = True,  metavar = 'chr:start:end:repeat_unit:max_size', type = str, help = 'first tandem repeat in the format of chr:start:end:repeat_unit:max_size. Positions start from 0. Start position is self-inclusive but end position is NOT self-inclusive (e.g. chr4:3074876:3074933:CAG:200)')
    parser.add_argument('-2', '--repeat2',      required = True,  metavar = 'chr:start:end:repeat_unit:max_size', type = str, help = 'second tandem repeat in the format of chr:start:end:repeat_unit:max_size. Positions start from 0. Start position is self-inclusive but end position is NOT self-inclusive (e.g. chr4:3074946:3074966:CCG:20)')
    parser.add_argument('-o', '--out_prefix',   required = True, metavar = 'path/to/out_dir/prefix_of_file_names',   type = str, help = '(required) prefix of output files.')
    

    ### optional arguments ### ploidy
    parser.add_argument('-d', '--data_type', required = False, metavar = 'data_type',  type = str, default = 'ont', help = '(required) sequencing data type. Should be one of the following: ont, ont_sup, ont_q20, clr, hifi')
    parser.add_argument('-c', '--num_threads',  required = False, metavar = 'INT',    type = int, default = 1,  help = 'number of threads used by minimap2 (default: 1)')
    parser.add_argument('--minimap2',     required = False, metavar = 'PATH',   type = str, default = '', help = 'path to minimap2 (default: using environment default)')
    parser.add_argument('--ploidy',       required = False, metavar = 'INT',    type = int, default = 2,  help = 'ploidy of the sample (default: 2)')
    parser.add_argument('--error_rate',   required = False, metavar = 'FLOAT',  type = float, default = 0.1,  help = 'sequencing error rate (default: 0.1)')
    parser.add_argument('--max_mutual_overlap', required = False, metavar = 'FLOAT',  type = float, default = 0.1,  help = 'max mutual overlap of two alleles in terms of repeat size distribution (default value: 0.1). If the Gaussian distribution of two alleles have more overlap than this value, the two alleles will be merged into one allele.')
    parser.add_argument('--remove_noisy_reads', required = False, action='store_true', help = 'remove noisy components when there are more components than ploidy')
    parser.add_argument('--max_num_components', required = False, metavar = 'INT',  type = int, default = -1,  help = 'max number of components for the Gaussian mixture model (default value: ploidy + 20). Some noisy reads and outlier reads may form a component. Therefore the number of components is usually larger than ploidy. If your sample have too many outlier reads, you can increase this number.')


    if len(sys.argv) < 2 or sys.argv[1] in ['help', 'h', '-help', 'usage']:
        input_args = parser.parse_args(['--help'])
    else:
        input_args = parser.parse_args()

    if input_args.data_type not in ['ont', 'ont_sup', 'ont_q20', 'clr', 'hifi']:
        tk.eprint(f'ERROR! data_type should be one of the following: ont, ont_sup, clr, hifi\n')
        tk.eprint(f'ont:     Oxford Nanopore sequencing, NOT Super Accuracy mode\n')
        tk.eprint(f'ont_sup: Oxford Nanopore sequencing, Super Accuracy mode\n')
        tk.eprint(f'ont_q20: Oxford Nanopore sequencing, Q20+ chemistry\n')
        tk.eprint(f'clr:     PacBio sequencing, Continuous Long Reads (CLR)\n')
        tk.eprint(f'hifi:    PacBio sequencing, HiFi/CCS reads\n')
        sys.exit(1)
        
    if input_args.ploidy < 1:
        tk.eprint('ERROR: --ploidy must be >= 1 !\n')
        sys.exit(1)

    if input_args.error_rate >= 1.0:
        tk.eprint('ERROR! --error_rate must be < 1\n')
        sys.exit(1)
    
    if input_args.max_mutual_overlap >= 1.0:
        tk.eprint('ERROR! --max_mutual_overlap must be < 1\n')
        sys.exit(1)

    if input_args.max_num_components == -1:
        input_args.max_num_components = input_args.ploidy + 20
    
    tk.check_input_file_exists(input_args.in_fq)
    tk.check_input_file_exists(input_args.ref_fasta)

    input_args.in_fq       = os.path.abspath(input_args.in_fq)
    input_args.ref_fasta   = os.path.abspath(input_args.ref_fasta)

    if input_args.minimap2 == '': 
        input_args.minimap2 = tk.find_executable_path('minimap2')
        if not input_args.minimap2:
            tk.eprint('ERROR! minimap2 was not found! Please supply the path to --minimap2')
            sys.exit(1)
        else:
            if os.path.exists(input_args.minimap2):
                tk.eprint('NOTICE: Found path to minimap2: %s' % input_args.minimap2)
            else:
                tk.eprint('ERROR! minimap2 was not found! Please supply the path to --minimap2')

    tk.check_input_file_exists(input_args.minimap2)

    if len(input_args.repeat1.split(':')) != 5:
        tk.eprint('ERROR! --repeat1 should be in this format: chr:start:end:repeat_unit:max_size (coordinates are 0-based, e.g. chr4:3074876:3074933:CAG:200)')
        sys.exit(1)
    
    if len(input_args.repeat2.split(':')) != 5:
        tk.eprint('ERROR! --repeat2 should be in this format: chr:start:end:repeat_unit:max_size (coordinates are 0-based, e.g. chr4:3074946:3074966:CCG:20)')
        sys.exit(1)

    return input_args

def main():

    input_args = parse_user_arguments()
    nanoRepeat_joint (input_args)

    return

def nanoRepeat_joint (input_args):

    n_input_reads = tk.count_fastq(input_args.in_fq)
    if n_input_reads < input_args.ploidy:
        tk.eprint(f'ERROR: No enough reads for analysis. Ploidy was set to {input_args.ploidy} but there were only {n_input_reads} reads in the fastq file: {input_args.in_fq}\n')
        sys.exit(1)
    
    out_dir, out_file_prefix = os.path.split(input_args.out_prefix)
    out_dir = os.path.abspath(out_dir)

    if out_file_prefix == '':
        out_file_prefix = os.path.split(input_args.in_fq)[1]
        input_args.out_prefix = os.path.join(out_dir, out_file_prefix)
    else:
        input_args.out_prefix = os.path.abspath(input_args.out_prefix)

    tk.create_dir(out_dir)

    tk.eprint(f'NOTICE: Input file is: {input_args.in_fq }')
    tk.eprint(f'NOTICE: Referece fasta file is: {input_args.ref_fasta}')
    tk.eprint(f'NOTICE: Repeat1 is: {input_args.repeat1}')
    tk.eprint(f'NOTICE: Repeat2 is: {input_args.repeat2}')
    tk.eprint(f'NOTICE: Output prefix is: {input_args.out_prefix}')
    

    repeat1 = Repeat().init_from_string(input_args.repeat1)
    repeat2 = Repeat().init_from_string(input_args.repeat2)

    if repeat1.chrom != repeat2.chrom:
        tk.eprint('ERROR: joint quantification only works with two nearby repeat regions. The two repeats in the config_file are in different chromosomes and there is no need to do joint quantification.\n')
        sys.exit(1)

    if repeat1.start > repeat2.start:
        repeat1, repeat2 = tk.switch_two_objects(repeat1, repeat2)

    repeat1.max_size += 10
    repeat2.max_size += 10
    max_anchor_len = 1000
    
    if repeat1.end + 100 < repeat2.start:
        tk.eprint('ERROR: joint quantification only works with two nearby repeat regions (distance < 100 bp). The two repeats are far away from each other and there is no need to do joint quantification.\n')
        sys.exit(1)

    temp_out_dir = os.path.join(out_dir, f'{out_file_prefix}.NanoRepeat.temp')
    tk.create_dir(temp_out_dir)

    repeat_chrom_seq = tk.read_one_chr_from_fasta_file(input_args.ref_fasta, repeat1.chrom)
    if len(repeat_chrom_seq) == 0:
        tk.eprint('ERROR: ref_fasta file: %s has no valid sequence!\n' % input_args.ref_fasta)
        sys.exit(1)

    initial_estimation = initial_estimate_repeat_size(input_args.minimap2, repeat_chrom_seq, input_args.in_fq, input_args.data_type, input_args.num_threads, repeat1, repeat2, max_anchor_len, temp_out_dir)

    final_estimation = fine_tune_read_count(initial_estimation, input_args.in_fq, repeat_chrom_seq, repeat1, repeat2, input_args.minimap2, input_args.data_type, input_args.num_threads, temp_out_dir)
    
    read_repeat_joint_count_dict = output_repeat_size_2d (input_args.in_fq, repeat1.repeat_id, repeat2.repeat_id, input_args.out_prefix, final_estimation.repeat1_count_dict, final_estimation.repeat2_count_dict)

    tk.eprint('NOTICE: Phasing reads using GMM')
    num_removed_reads = 0
    split_alleles_using_gmm_2d (input_args.ploidy, input_args.error_rate, input_args.max_mutual_overlap, input_args.remove_noisy_reads, input_args.max_num_components, repeat1, repeat2, read_repeat_joint_count_dict, num_removed_reads, input_args.in_fq, input_args.out_prefix)

    shutil.rmtree(temp_out_dir)
    tk.eprint(f'NOTICE: Program finished. Output files are here: {out_dir}\n')

    return


def fine_tune_read_count(initial_estimation, in_fastq_file, repeat_chrom_seq, repeat1, repeat2, minimap2, data_type, num_threads, out_dir):

    assert repeat1.chrom == repeat2.chrom
    assert repeat1.start < repeat2.start

    repeat1.round1_max_size = 0
    repeat2.round1_max_size = 0
    repeat1.round1_min_size = repeat1.max_size
    repeat2.round1_min_size = repeat2.max_size
    for readname, value in initial_estimation.repeat1_count_range_dict.items():
        min_repeat_size, max_repeat_size = value
        if max_repeat_size > repeat1.round1_max_size:
            repeat1.round1_max_size = max_repeat_size
        if min_repeat_size < repeat1.round1_min_size:
            repeat1.round1_min_size = min_repeat_size

    for readname, value in initial_estimation.repeat2_count_range_dict.items():
        min_repeat_size, max_repeat_size = value
        if max_repeat_size > repeat2.round1_max_size:
            repeat2.round1_max_size = max_repeat_size
        if min_repeat_size < repeat2.round1_min_size:
            repeat2.round1_min_size = min_repeat_size

    
    if repeat1.round1_max_size > repeat1.max_size: repeat1.round1_max_size = repeat1.max_size
    if repeat2.round1_max_size > repeat2.max_size: repeat2.round1_max_size = repeat2.max_size

    tk.eprint('NOTICE: In round 1 estimation, repeat 1 (%s) is in the range of (%d, %d)' % (repeat1.repeat_unit, repeat1.round1_min_size, repeat1.round1_max_size))
    tk.eprint('NOTICE: In round 1 estimation, repeat 2 (%s) is in the range of (%d, %d)' % (repeat2.repeat_unit, repeat2.round1_min_size, repeat2.round1_max_size))
    
    fastq_dict = fastq_file_to_dict(in_fastq_file)

    round2_estimation = round2_estimation_of_repeat_size(initial_estimation, fastq_dict, repeat_chrom_seq, repeat1, repeat2, minimap2, data_type, num_threads, out_dir)
    
    if round2_estimation.step_size1 > 1 and round2_estimation.step_size2 > 1:
        final_estimation = round3_estimation_of_repeat_size(initial_estimation, round2_estimation, fastq_dict, repeat_chrom_seq, repeat1, repeat2, minimap2, data_type, num_threads, out_dir)
    else:
        final_estimation = round2_estimation
    
    return final_estimation

def round3_estimation_of_repeat_size(initial_estimation, round2_estimation, fastq_dict, repeat_chrom_seq, repeat1, repeat2, minimap2, data_type, num_threads, out_dir):

    if len(round2_estimation.repeat1_count_dict) == 0 or len(round2_estimation.repeat2_count_dict) == 0:
        return RepeatSize()
    
    assert repeat1.chrom == repeat2.chrom
    assert repeat1.start < repeat2.start
    max_flanking_len = 1000

    buffer_size1 = round2_estimation.step_size1
    buffer_size2 = round2_estimation.step_size2
    tk.eprint(f'NOTICE: Fine-tuning repeat size. buffer_size1 = {buffer_size1}, buffer_size2 = {buffer_size2}')

    round2_size1_list = list()
    round2_size2_list = list()

    for readname in round2_estimation.repeat1_count_dict:
        if readname not in round2_estimation.repeat2_count_dict: continue
        size1 = round2_estimation.repeat1_count_dict[readname]
        size2 = round2_estimation.repeat2_count_dict[readname]
        round2_size1_list.append(size1)
        round2_size2_list.append(size2)

    min_size1 = int(min(round2_size1_list) - buffer_size1)
    max_size1 = int(max(round2_size1_list) + buffer_size1 + 2)
    if min_size1 < 0: min_size1 = 0
    min_size2 = int(min(round2_size2_list) - buffer_size2)
    max_size2 = int(max(round2_size2_list) + buffer_size2 + 2)
    if min_size2 < 0: min_size2 = 0

    ## round 3 alignment ##
    left_anchor_seq, mid_anchor_seq, right_anchor_seq = extract_anchor_seq_for_two_repeats (repeat_chrom_seq, repeat1, repeat2, max_flanking_len)
    left_anchor_len  = len(left_anchor_seq)
    mid_anchor_len   = len(mid_anchor_seq)

    preset = tk.get_preset_for_minimap2(data_type)
    round3_paf_file =  os.path.join(out_dir, 'round3.paf')
    round3_paf_f = open(round3_paf_file, 'w')
    round3_paf_f.close()

    for repeat_count1 in range(min_size1, max_size1):
        for repeat_count2 in range(min_size2, max_size2):
            tmp_fastq_file = os.path.join(out_dir, '%d.%d.round3.fastq' % (repeat_count1, repeat_count2) )
            tmp_fastq_f = open(tmp_fastq_file, 'w')
            n_tmp_reads = 0
            for readname in fastq_dict:
                if readname not in round2_estimation.repeat1_count_dict: continue
                if readname not in round2_estimation.repeat2_count_dict: continue
                size1 = round2_estimation.repeat1_count_dict[readname]
                size2 = round2_estimation.repeat2_count_dict[readname]
                if repeat_count1 < size1 - buffer_size1 or repeat_count1 >= size1 + buffer_size1: continue
                if repeat_count2 < size2 - buffer_size2 or repeat_count2 >= size2 + buffer_size2: continue
                r1min1, r1max1 = initial_estimation.repeat1_count_range_dict[readname]
                r1min2, r1max2 = initial_estimation.repeat2_count_range_dict[readname]
                if repeat_count1 < r1min1 or repeat_count1 >= r1max1: continue
                if repeat_count2 < r1min2 or repeat_count2 >= r1max2: continue

                tmp_fastq_f.write(fastq_dict[readname])
                n_tmp_reads += 1
            tmp_fastq_f.close()
            if n_tmp_reads == 0: 
                tk.rm(tmp_fastq_file)
                continue
            tmp_ref_file = os.path.join(out_dir, '%d.%d.ref.fasta' % (repeat_count1, repeat_count2))
            build_fasta_template_for_two_repeats(left_anchor_seq, mid_anchor_seq, right_anchor_seq, repeat1, repeat2, repeat_count1, repeat_count2, tmp_ref_file)
            cmd = f'{minimap2} -c --eqx -t {num_threads} {preset} {tmp_ref_file} {tmp_fastq_file} >> {round3_paf_file} 2> /dev/null'
            tk.run_system_cmd(cmd)
            tk.rm(tmp_ref_file)
            tk.rm(tmp_fastq_file)

    round3_estimation = estimate_two_repeats_from_paf(round3_paf_file, left_anchor_len, mid_anchor_len, repeat1, repeat2)
    round3_estimation.step_size1 = 1
    round3_estimation.step_size2 = 1

    return round3_estimation

def choose_best_step_size(repeat, count_range_dict):

    max_len = 50
    max_step_size = int(max_len/repeat.repeat_unit_size)
    if max_step_size < 1: max_step_size = 1

    error_list = list()
    for readname in count_range_dict:
        a, b = count_range_dict[readname]
        error = b - a
        error_list.append(error)

    l = np.mean(error_list)

    count_list = list()
    for size in range(1, max_step_size+1):
        count = int(l/size)+1
        count += size * 2 + 2
        tup = (size, count)
        count_list.append(tup)

    count_list.sort(key = lambda x:x[1])
  
    return count_list[0][0]

def round2_estimation_of_repeat_size(initial_estimation, fastq_dict, repeat_chrom_seq, repeat1, repeat2, minimap2, data_type, num_threads, out_dir):
    
    assert repeat1.chrom == repeat2.chrom
    assert repeat1.start < repeat2.start
    max_flanking_len = 1000

    step_size1 = choose_best_step_size(repeat1, initial_estimation.repeat1_count_range_dict)
    step_size2 = choose_best_step_size(repeat2, initial_estimation.repeat2_count_range_dict)
    
    tk.eprint(f'NOTICE: Round 2 estimation. step_size1 = {step_size1}; step_size2 = {step_size2}')

    ## round 2 alignment ##
    left_anchor_seq, mid_anchor_seq, right_anchor_seq = extract_anchor_seq_for_two_repeats (repeat_chrom_seq, repeat1, repeat2, max_flanking_len)
    left_anchor_len  = len(left_anchor_seq)
    mid_anchor_len   = len(mid_anchor_seq)

    preset = tk.get_preset_for_minimap2(data_type)
    round2_paf_file =  os.path.join(out_dir, 'round2.paf')
    round2_paf_f = open(round2_paf_file, 'w')
    round2_paf_f.close()

    for repeat_count1 in range(repeat1.round1_min_size, repeat1.round1_max_size + 1, step_size1):
        for repeat_count2 in range(repeat2.round1_min_size, repeat2.round1_max_size + 1, step_size2):
            tmp_fastq_file = os.path.join(out_dir, '%d.%d.round2.fastq' % (repeat_count1, repeat_count2) )
            tmp_fastq_f = open(tmp_fastq_file, 'w')
            n_tmp_reads = 0
            for readname in fastq_dict:
                if readname not in initial_estimation.repeat1_count_range_dict: continue
                if readname not in initial_estimation.repeat2_count_range_dict: continue
                min1, max1 = initial_estimation.repeat1_count_range_dict[readname]
                min2, max2 = initial_estimation.repeat2_count_range_dict[readname]
                if repeat_count1 >= min1 and repeat_count1 < max1 and repeat_count2 >= min2 and repeat_count2 < max2:                
                    tmp_fastq_f.write(fastq_dict[readname])
                    n_tmp_reads += 1
            tmp_fastq_f.close()
            if n_tmp_reads == 0:
                tk.rm(tmp_fastq_file)
                continue
            tmp_ref_file = os.path.join(out_dir, '%d.%d.ref.fasta' % (repeat_count1, repeat_count2))
            build_fasta_template_for_two_repeats(left_anchor_seq, mid_anchor_seq, right_anchor_seq, repeat1, repeat2, repeat_count1, repeat_count2, tmp_ref_file)
            cmd = f'{minimap2} {preset} -c --eqx -t {num_threads}  {tmp_ref_file} {tmp_fastq_file} >> {round2_paf_file} 2> /dev/null'
            tk.run_system_cmd(cmd)
            tk.rm(tmp_ref_file)
            tk.rm(tmp_fastq_file)

    round2_estimation = estimate_two_repeats_from_paf(round2_paf_file, left_anchor_len, mid_anchor_len, repeat1, repeat2)
    round2_estimation.step_size1 = step_size1
    round2_estimation.step_size2 = step_size2

    return round2_estimation

def estimate_two_repeats_from_paf(in_paf_file, left_anchor_len, mid_anchor_len, repeat1, repeat2):

    repeat_size_estimation = RepeatSize()

    in_paf_f = open(in_paf_file, 'r')
    read_align_score_dict = dict()
    while 1:
        line = in_paf_f.readline()
        if not line: break
        line = line.strip()
        if not line: continue

        col_list = line.strip().split('\t')
        paf = PAF(col_list)
        repeat_size1, repeat_size2 = paf.tname.split('-')
        repeat_size1 = int(repeat_size1)
        repeat_size2 = int(repeat_size2)

        a = left_anchor_len - 10
        if a < 0: a = 0
        b = left_anchor_len + repeat1.repeat_unit_size * repeat_size1 + mid_anchor_len + repeat2.repeat_unit_size * repeat_size2 + 10
        if b > paf.tlen: b = paf.tlen
        repeat_region_align_score = tk.target_region_alignment_stats_from_cigar(paf.cigar, paf.tstart, paf.tend, a, b).score
    
        tup = (repeat_size1, repeat_size2, repeat_region_align_score, paf.align_score, paf.qlen)
        if paf.qname not in read_align_score_dict:
            read_align_score_dict[paf.qname] = list()
        read_align_score_dict[paf.qname].append(tup)    

    in_paf_f.close()

    for readname in read_align_score_dict:
        tup_list = read_align_score_dict[readname]
        tup_list.sort(key = lambda x:x[2], reverse = True)
        max_score = tup_list[0][2]
        max_score_size1_list = list()
        max_score_size2_list = list()
        for tup in tup_list:
            if tup[2] == max_score:
                size1 = tup[0]
                size2 = tup[1]
                max_score_size1_list.append(size1)
                max_score_size2_list.append(size2)
            else:
                break
        
        round2_repeat_size1 = np.mean(max_score_size1_list)
        round2_repeat_size2 = np.mean(max_score_size2_list)
        repeat_size_estimation.repeat1_count_dict[readname] = round2_repeat_size1
        repeat_size_estimation.repeat2_count_dict[readname] = round2_repeat_size2
    
    return repeat_size_estimation

def extract_anchor_seq_for_two_repeats (repeat_chrom_seq, repeat1, repeat2, max_flanking_len):

    left_end_pos = repeat1.start
    left_start_pos = left_end_pos - max_flanking_len
    if left_start_pos < 0: left_start_pos = 0
    
    mid_start_pos = repeat1.end
    mid_end_pos = repeat2.start

    right_start_pos = repeat2.end
    right_end_pos = right_start_pos + max_flanking_len
    if right_end_pos > len(repeat_chrom_seq): right_end_pos = len(repeat_chrom_seq)

    left_anchor_seq  = repeat_chrom_seq[left_start_pos:left_end_pos]
    mid_anchor_seq   = repeat_chrom_seq[mid_start_pos:mid_end_pos]
    right_anchor_seq = repeat_chrom_seq[right_start_pos:right_end_pos]

    return left_anchor_seq, mid_anchor_seq, right_anchor_seq

def build_fasta_template_for_two_repeats(left_anchor_seq, mid_anchor_seq, right_anchor_seq, repeat1, repeat2, repeat_count1, repeat_count2, tmp_ref_file):

    tmp_ref_f = open(tmp_ref_file, 'w')
    tmp_ref_f.write('>%d-%d\n' % (repeat_count1, repeat_count2))
    seq = left_anchor_seq + repeat1.repeat_unit * repeat_count1 + mid_anchor_seq + repeat2.repeat_unit * repeat_count2 + right_anchor_seq + '\n'
    tmp_ref_f.write(seq)
    tmp_ref_f.close()

    return

def initial_estimate_repeat_size(minimap2, repeat_chrom_seq, in_fastq_file, data_type, num_threads, repeat1, repeat2, max_anchor_len, out_dir):

    tk.eprint('NOTICE: Round 1 estimation')
    assert repeat1.chrom == repeat2.chrom
    assert repeat1.start < repeat2.start

    left_anchor_seq, right_anchor_seq = tk.extract_anchor_sequence(repeat_chrom_seq, repeat1.start, repeat2.end, max_anchor_len)
    left_anchor_len  = len(left_anchor_seq)
    right_anchor_len = len(right_anchor_seq)

    ## make template fasta file ##
    left_template_fasta_file = os.path.join(out_dir, 'round1_templates.left_anchor.fasta')
    left_template_name  = 'left_anchor_%d_%d_%s' % (len(left_anchor_seq), repeat1.max_size, repeat1.repeat_unit)
    left_template_seq   = left_anchor_seq + repeat1.repeat_unit * repeat1.max_size

    left_template_fasta_f = open(left_template_fasta_file, 'w')
    left_template_fasta_f.write('>%s\n' % left_template_name)
    left_template_fasta_f.write('%s\n' % left_template_seq)
    left_template_fasta_f.close()


    right_template_fasta_file = os.path.join(out_dir, 'round1_templates.right_anchor.fasta')
    right_template_name = 'right_anchor_%d_%d_%s_revc' % (len(right_anchor_seq), repeat2.max_size, repeat2.repeat_unit)
    right_template_seq  = repeat2.repeat_unit * repeat2.max_size + right_anchor_seq
    right_template_seq  = tk.rev_comp(right_template_seq)
    
    right_template_fasta_f = open(right_template_fasta_file, 'w')
    right_template_fasta_f.write('>%s\n' % right_template_name)
    right_template_fasta_f.write('%s\n' % right_template_seq)
    right_template_fasta_f.close()

    round1_paf_file = os.path.join(out_dir, 'round1.paf')
    round1_left_anchor_paf_file  = os.path.join(out_dir, 'round1.left.paf')
    round1_right_anchor_paf_file = os.path.join(out_dir, 'round1.right.paf')
    preset = tk.get_preset_for_minimap2(data_type)

    cmd = f'{minimap2} {preset} -c --eqx -t {num_threads} {left_template_fasta_file} {in_fastq_file} > {round1_left_anchor_paf_file} 2> /dev/null'
    tk.run_system_cmd(cmd)
    
    cmd = f'{minimap2} {preset} -c --eqx -t {num_threads} {right_template_fasta_file} {in_fastq_file} > {round1_right_anchor_paf_file} 2> /dev/null'
    tk.run_system_cmd(cmd)

    cmd = f'cat {round1_left_anchor_paf_file} {round1_right_anchor_paf_file} | sort -k1 - > {round1_paf_file}'
    tk.run_system_cmd(cmd)

    tk.rm(round1_left_anchor_paf_file)
    tk.rm(round1_right_anchor_paf_file)
    initial_estimation = round1_estimation_from_paf(round1_paf_file, repeat1, repeat2, left_anchor_len, right_anchor_len)

    return initial_estimation

def round1_estimation_from_paf(round1_paf_file, repeat1, repeat2, left_anchor_len, right_anchor_len):

    initial_estimation = Round1Estimation()
    round1_paf_f = open(round1_paf_file, 'r')
    read_paf_list = list()
    while 1:
        line = round1_paf_f.readline()
        if not line: break
        line = line.strip()
        if not line: continue

        col_list = line.strip().split('\t')
        paf = PAF(col_list)

        if len(read_paf_list) == 0 or paf.qname == read_paf_list[0].qname:
            read_paf_list.append(paf)
        else:
            round1_estimation_for1read(read_paf_list, repeat1, repeat2, left_anchor_len, right_anchor_len, initial_estimation)
            read_paf_list.clear()
            read_paf_list.append(paf)

    round1_paf_f.close()
    
    round1_estimation_for1read(read_paf_list, repeat1, repeat2, left_anchor_len, right_anchor_len, initial_estimation)

    for readname in initial_estimation.bad_reads_set:
        initial_estimation.repeat1_count_range_dict.pop(readname, None)
        initial_estimation.repeat2_count_range_dict.pop(readname, None)

    return initial_estimation
            

def round1_estimation_for1read(read_paf_list, repeat1, repeat2, left_anchor_len, right_anchor_len, initial_estimation):

    left_boundary_pos  = left_anchor_len
    right_boundary_pos = right_anchor_len
    
    left_anchor_paf_list  = list()
    right_anchor_paf_list = list()
    
    for paf in read_paf_list:
        if paf.mapq < 30: continue
        if paf.is_primary == False: continue 
        if paf.tname[0:5] == 'left_':
            if paf.tstart <= left_boundary_pos and paf.tend >= left_boundary_pos:
                left_anchor_paf_list.append(paf)
        elif paf.tname[0:5] == 'right':
            if paf.tstart <= right_boundary_pos and paf.tend >= right_boundary_pos:
                right_anchor_paf_list.append(paf)
        else:
            tk.eprint('ERROR! unknown template: %s' % (paf.tname))
            sys.exit(1)
    
    if len(left_anchor_paf_list) != 1 or len(right_anchor_paf_list) != 1 : return
    left_paf  = left_anchor_paf_list[0]
    right_paf  = right_anchor_paf_list[0]

    if left_paf.strand == right_paf.strand: return

    readname = left_paf.qname
    left_paf_max_repeat_size  = int((left_paf.tend - left_boundary_pos)/repeat1.repeat_unit_size) + 5
    left_paf_min_repeat_size  = tk.calculate_repeat_size_from_exact_match(left_paf.cigar, left_paf.tstart,   left_boundary_pos,  repeat1.repeat_unit_size)

    a = max(0, left_paf_min_repeat_size - 20)
    b = int(left_paf_min_repeat_size / 2.0)
    left_paf_min_repeat_size = min(a, b)
    

    initial_estimation.repeat1_count_range_dict[readname] = (left_paf_min_repeat_size, left_paf_max_repeat_size)

    
    right_paf_max_repeat_size = int((right_paf.tend - right_boundary_pos)/repeat2.repeat_unit_size) + 5
    right_paf_min_repeat_size = tk.calculate_repeat_size_from_exact_match(right_paf.cigar, right_paf.tstart, right_boundary_pos, repeat2.repeat_unit_size)
    a = max(0, right_paf_min_repeat_size - 20)
    b = int(right_paf_min_repeat_size / 2.0)
    right_paf_min_repeat_size = min(a, b)
    
    initial_estimation.repeat2_count_range_dict[readname] = (right_paf_min_repeat_size, right_paf_max_repeat_size)

    if left_paf.strand == '+': 
        candidate_start = left_paf.qstart
        candidate_end = right_paf.qlen - right_paf.qstart
        if candidate_end - candidate_start <= 0: initial_estimation.bad_reads_set.add(readname)
    else:
        candidate_start = right_paf.qstart
        candidate_end = left_paf.qlen - left_paf.qstart
        if candidate_end - candidate_start <= 0: initial_estimation.bad_reads_set.add(readname)

    initial_estimation.potential_repeat_region_dict[readname] = (candidate_start, candidate_end)
    return


def fastq_file_to_dict(in_fastq_file):

    fastq_dict = dict()

    in_fastq_f = tk.gzopen(in_fastq_file)
    while 1:
        line1 = in_fastq_f.readline()
        line2 = in_fastq_f.readline()
        line3 = in_fastq_f.readline()
        line4 = in_fastq_f.readline()

        if not line1: break
        if not line2: break
        if not line3: break
        if not line4: break

        readname = line1.strip().split()[0][1:]
        fastq_dict[readname] = line1 + line2 + line3 + line4

    in_fastq_f.close()

    return fastq_dict

def remove_noisy_reads_2d(allele_list, ploidy, error_rate, max_mutual_overlap, max_num_components, repeat1, repeat2, in_fastq_file, out_prefix):
    tk.eprint('NOTICE: Try to remove noisy reads')
    allele_list.sort(key = lambda allele:allele.num_reads)
    num_removed_reads = 0
    while len(allele_list) > ploidy and len(allele_list) >= 2:
        if allele_list[0].num_reads * 1.5 <= allele_list[-ploidy].num_reads:
            num_removed_reads += allele_list[0].num_reads
            allele_list.pop(0)
        else:
            break

    read_repeat_joint_count_dict = dict()
    tk.eprint(f'NOTICE: There are {len(allele_list)} alleles after removing noisy reads')
    for allele in allele_list:
        for i in range(0, len(allele.readname_list)):
            readname = allele.readname_list[i]
            repeat_size1 = allele.repeat1_size_list[i]
            repeat_size2 = allele.repeat2_size_list[i]
            read_repeat_joint_count_dict[readname] = (repeat_size1, repeat_size2)
    
    remove_noisy_reads = False
    return split_alleles_using_gmm_2d (ploidy, error_rate, max_mutual_overlap, remove_noisy_reads, max_num_components, repeat1, repeat2, read_repeat_joint_count_dict, num_removed_reads, in_fastq_file, out_prefix)


def split_alleles_using_gmm_2d (ploidy, error_rate, max_mutual_overlap, remove_noisy_reads, max_num_components, repeat1, repeat2, read_repeat_joint_count_dict, num_removed_reads, in_fastq_file, out_prefix):
    
    if ploidy < 1:
        tk.eprint('ploidy must be >= 1 !\n')
        sys.exit(1)

    if len(read_repeat_joint_count_dict) < ploidy or len(read_repeat_joint_count_dict) == 1:
        tk.eprint('WARNING: No enough reads! input fastq file is: %s\n' % in_fastq_file)
        return

    cov_type = 'diag'
    dimension = 2
    probability_cutoff = 0.95

    readname_list, read_repeat_count_list = remove_outlier_reads_2d(read_repeat_joint_count_dict)
    read_repeat_count_array = np.array(read_repeat_count_list).reshape(-1, dimension)
    simulated_read_repeat_count_list = simulate_reads(read_repeat_count_list, error_rate)
    simualted_read_repeat_count_array = np.array(simulated_read_repeat_count_list).reshape(-1, dimension)

    best_n_components, final_gmm = auto_GMM_2d (simualted_read_repeat_count_array, max_num_components, cov_type, max_mutual_overlap)
    tk.eprint('NOTICE: Number of alleles = %d' % best_n_components)
    
    allele_list = create_allele_list_2d(best_n_components, final_gmm, readname_list, read_repeat_count_array, read_repeat_joint_count_dict, probability_cutoff)
    
    num_removed_reads = 0
    if remove_noisy_reads == True and len(allele_list) > ploidy:
        return remove_noisy_reads_2d(allele_list, ploidy, error_rate, max_mutual_overlap, max_num_components, repeat1, repeat2, in_fastq_file, out_prefix)
    
    allele_list.sort(key = lambda allele:allele.gmm_mean1)

    readinfo_dict = create_readinfo_dict_from_allele_list(allele_list, dimension)

    score_cut_off = calculate_log_likelyhood_cutoff(final_gmm, 0.95)

    tk.eprint('NOTICE: Writing phasing results...')
    output_phasing_results_2d(allele_list, repeat1.repeat_id, repeat2.repeat_id, in_fastq_file, out_prefix)

    tk.eprint('NOTICE: Writing to output fastq files...')
    output_phased_fastq(in_fastq_file, readinfo_dict, best_n_components, out_prefix)

    tk.eprint('NOTICE: Writing summary file...')
    output_summary_file_2d(in_fastq_file, allele_list, repeat1.repeat_id, repeat2.repeat_id, num_removed_reads, out_prefix)

    tk.eprint('NOTICE: Plotting figures...')
    plot_repeat_counts_2d(readinfo_dict, allele_list, repeat1.repeat_id, repeat2.repeat_id, out_prefix)

    scatter_plot_with_contour_2d (read_repeat_joint_count_dict, final_gmm, score_cut_off, repeat1.repeat_id, repeat2.repeat_id, allele_list, out_prefix)

    return


def calculate_log_likelyhood_cutoff(gmm, ci):
    X, labels = gmm.sample(100000)
    score_list = gmm.score_samples(X)
    X_score_list = list()
    for i in range(0, len(X)):
        X_score = [X[i], score_list[i]]
        X_score_list.append(X_score)

    X_score_list.sort(key = lambda x:x[1], reverse = True)
    cut_off_idx = int(len(X_score_list) * float(ci))
    log_likelyhood_cutoff = X_score_list[cut_off_idx][1]

    return log_likelyhood_cutoff


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

'''
Copyright (c) 2020-2023 Children's Hospital of Philadelphias
Author: Li Fang (fangli2718@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import os
import sys

## objects ##
class Read:
    def __init__(self):
        self.read_name = None
        self.full_read_len = None
        self.core_seq = None
        self.init_repeat_size = None
        self.left_anchor_is_good = False
        self.right_anchor_is_good = False
        self.both_anchors_are_good = False
        self.core_seq_start_pos = None
        self.core_seq_end_pos = None
        self.mid_seq_start_pos = None
        self.mid_seq_end_pos = None
        self.left_anchor_paf = None
        self.right_anchor_paf = None
        self.dist_between_anchors = None
        self.seq_between_anchors = None
        self.strand = None
        self.left_buffer_len = None
        self.right_buffer_len = None
        self.round1_repeat_size = None
        self.round2_repeat_size = None
        self.round3_repeat_size = None

class RepeatRegion:
    def __init__(self, line = None):
        self.left_anchor_seq = None
        self.right_anchor_seq = None
        
        self.left_anchor_len = None
        self.right_anchor_len = None

        self.anchor_len = None
        self.mid_ref_seq = None

        self.region_fq_file = None
        self.region_fasta_file = None # template
        self.core_seq_fq_file = None
        self.mid_seq_fq_file = None
        self.chrom = None
        self.start_pos = None
        self.end_pos = None
        self.repeat_unit_seq = None
        self.max_repeat_size = None

        self.temp_out_dir = None
        self.final_out_dir = None
        self.out_prefix = None

        self.temp_file_list = []
        self.ref_has_issue = False
        self.read_dict = dict()
        self.read_core_seq_dict = dict()
        self.buffer_len = None
        self.brute_force_repeat_count_dict = dict()
        
    
        if line != None:
            col_list = line.strip().split('\t')
            if len(col_list) < 4:
                sys.stderr.write('ERROR! the repeat region bed file should be tab-delimited and have 4 columns: chrom, start_position, end_position, repeat_unit. start_position and end_position should be 0-based')
                sys.exit(1)
            
            self.chrom, self.start_pos, self.end_pos, self.repeat_unit_seq = col_list[0:4]
            self.start_pos = int(self.start_pos)
            self.end_pos = int(self.end_pos)
    
    def to_invertal(self, flank_dist = 0):
        assert (flank_dist >= 0)
        start_pos = self.start_pos - flank_dist
        end_pos = self.end_pos + flank_dist
        if start_pos < 0: start_pos = 0
        return f'{self.chrom}:{start_pos}-{end_pos}'

    def to_unique_id(self):
        return f'{self.chrom}-{self.start_pos}-{self.end_pos}-{self.repeat_unit_seq}'

    def to_outfile_prefix(self):
        if len(self.repeat_unit_seq) < 30:
            seq = self.repeat_unit_seq
        else:
            seq = self.repeat_unit_seq[0:20] + '....' + self.repeat_unit_seq[-6:]

        return f'{self.chrom}-{self.start_pos}-{self.end_pos}-{seq}'
    
def read_repeat_region_file(repeat_region_file):
    repeat_region_list = []
    repeat_region_f = open(repeat_region_file, 'r')
    lines = list(repeat_region_f)
    repeat_region_f.close()
    for line in lines:
        repeat_region = RepeatRegion(line)
        repeat_region_list.append(repeat_region)

    return repeat_region_list



from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()
    
setup(
    name='NanoRepeat',
    version='1.7',
    description='NanoRepeat: quantification of Short Tandem Repeats (STRs) from long-read sequencing data',
    long_description = readme,
    long_description_content_type = 'text/markdown',
    url='https://github.com/WGLab/NanoRepeat',
    author='Li Fang, Kai Wang',
    author_email='fangli9@sysu.edu.cn',
    license='MIT',
    packages = find_packages("src"),
    package_dir = {"": "src"},
    data_files = [("", ["LICENSE"])],
    scripts=['src/NanoRepeat/nanoRepeat.py', 'src/NanoRepeat/nanoRepeat-joint.py'],
    install_requires=['matplotlib>=3.4.0',
                      'numpy>=1.21.6', 
                      'scikit-learn>=0.22.1']
)



#!/usr/bin/env python3

'''
Copyright (c) 2020-2023 Children's Hospital of Philadelphia
Author: Li Fang (fangli2718@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import os
import sys

import gzip
from datetime import datetime
import distutils
from distutils import spawn
from timeit import repeat

TimeFormat = '%m/%d/%Y %H:%M:%S'

class ReadError:
    def __init__(self, num_match = 0, num_mismatch = 1000000, num_ins = 1000000, num_del = 1000000, align_score = 0):
        self.num_match = num_match
        self.num_mismatch = num_mismatch
        self.num_ins = num_ins
        self.num_del = num_del
        self.score = align_score
    
    def num_edit_bases(self):
        return self.num_mismatch + self.num_ins + self.num_del

### IO ###

def read_list_file(in_list_file, abspath = False):
    out_list = list()
    in_list_fp = open(in_list_file, 'r')
    lines = list(in_list_fp)
    in_list_fp.close()

    for line in lines:
        line = line.strip()
        if abspath == True:
            line = os.path.abspath(line)
            out_list.append(line)
        else:
            out_list.append(line)
    return out_list

def gzopen(in_file, mode = 'rt'):

    if '.gz' == in_file[-3:] or '.bgz' == in_file[-4:]:
        in_fp = gzip.open(in_file, 'rt')
    else:
        in_fp = open(in_file, 'rt')

    return in_fp

def create_dir(dir):

    os.makedirs(dir, exist_ok=True)

    return

def check_input_file_exists(in_file):

    if os.path.exists(in_file) == False:
        eprint('ERROR! the file was not found: %s' % in_file)
        sys.exit()

    return

def check_output_file_exists(out_file):

    if os.path.exists(out_file) == False or os.path.getsize(out_file) == 0:
        return False
    else:
        return True

def eprint(message):
    sys.stderr.write('[' + datetime.now().strftime(TimeFormat) + '] ' + message + '\n')
    return

def get_file_prefix(input_file):

    return os.path.splitext(os.path.split(input_file)[1])[0]

### FASTQ/FASTA ###
    
def count_fastq (in_fastq_file):

    n_reads = 0

    in_fastq_fp = gzopen(in_fastq_file)
    while 1:
        line1 = in_fastq_fp.readline()
        line2 = in_fastq_fp.readline()
        line3 = in_fastq_fp.readline()
        line4 = in_fastq_fp.readline()

        if not line1: break
        if not line2: break
        if not line3: break
        if not line4: break

        n_reads += 1

    in_fastq_fp.close()

    return n_reads

def fasta_file2dict(fasta_file):

    fasta_fp = gzopen(fasta_file)
    fasta_dict = dict()

    curr_name = ''
    curr_seq = ''

    while 1:
        line = fasta_fp.readline()
        if not line: break
        line = line.strip()
        if not line: continue 
        if line[0] ==  '>':
            if len(curr_seq) > 0 and len(curr_name) > 0:
                assert curr_name not in fasta_dict
                fasta_dict[curr_name] = curr_seq
            curr_name = line[1:].split()[0]
            curr_seq = ''
            continue

        curr_seq += line
     
    fasta_fp.close()

    if len(curr_seq) > 0 and len(curr_name) > 0:
        fasta_dict[curr_name] = curr_seq
    
    return fasta_dict

def read_fasta_file(fasta_file):

    fasta_fp = gzopen(fasta_file)

    fasta_name_list = list()
    fasta_seq_list = list()
    curr_name = ''
    curr_seq = ''

    while 1:
        line = fasta_fp.readline()
        if not line: break
        line = line.strip()
        if not line: continue 
        if line[0] ==  '>':
            if len(curr_seq) > 0 and len(curr_name) > 0:
                fasta_name_list.append(curr_name)
                fasta_seq_list.append(curr_seq)
            curr_name = line[1:].split()[0]
            curr_seq = ''
            continue

        curr_seq += line
     
    fasta_fp.close()

    if len(curr_seq) > 0 and len(curr_name) > 0:
        fasta_name_list.append(curr_name)
        fasta_seq_list.append(curr_seq)
    
    return fasta_name_list, fasta_seq_list


def read_one_chr_from_fasta_file(fasta_file, target_chr):

    if '.gz' == fasta_file[-3:]:
        fasta_fp = gzip.open(fasta_file, 'rt')
    else:
        fasta_fp = open(fasta_file, 'rt')

    target_seq = ''

    status = 's' # skip
    while 1:
        line = fasta_fp.readline()
        if not line: break
        line = line.strip()
        if not line: continue
        if line[0] ==  '>':
            new_contig_name = line[1:].split()[0]
            if new_contig_name == target_chr:
                status = 'r'
                continue
            else:
                if len(target_seq) > 0:
                    status = 'd'
                    break
                else:
                    status = 's'
                    continue

        if status == 'd': 
            break
        elif status == 's':
            continue
        elif status == 'r':
            target_seq += line
     
    fasta_fp.close()
    
    return target_seq


def extract_anchor_sequence(repeat_chrom_seq, repeat_start, repeat_end, max_anchor_len):
    
    anchor_start = repeat_start - max_anchor_len
    anchor_end = repeat_start
    if anchor_start < 0: anchor_start = 0
    left_anchor_seq = repeat_chrom_seq[anchor_start:anchor_end]

    anchor_start = repeat_end
    anchor_end = anchor_start + max_anchor_len
    if anchor_end > len(repeat_chrom_seq):
        anchor_end = len(repeat_chrom_seq)
    right_anchor_seq = repeat_chrom_seq[anchor_start:anchor_end]
    
    return left_anchor_seq, right_anchor_seq

def split_fastq(in_fastq_file_list, num_out_file, out_prefix):

    out_readname_set = set() # used to skip duplicated reads
    out_fastq_file_list = list()
    out_fastq_fp_list = list()
    for i in range(0, num_out_file):
        out_fastq_file = out_prefix + '.part%d.fastq' % i
        out_fastq_file_list.append(out_fastq_file)
        out_fastq_fp = open(out_fastq_file, 'w')
        out_fastq_fp_list.append(out_fastq_fp)

    for in_fastq_file in in_fastq_file_list:
        in_fastq_fp = gzopen(in_fastq_file)

        i = 0
        status = 0
        while 1:
            if status == 0:
                line1 = in_fastq_fp.readline()
            line2 = in_fastq_fp.readline()
            line3 = in_fastq_fp.readline()
            line4 = in_fastq_fp.readline()
            if not line1: break
            if not line2: break
            if not line3: break
            if not line4: break

            if line1[0] != '@' or len(line2) != len(line4) or line3.strip() != '+':
                eprint('ERROR! Bad fastq file: %s' % in_fastq_file)
                break
                

            i += 1
            file_id = i % num_out_file
            readname = line1.strip().split()[0][1:]
            if readname not in out_readname_set:
                out_fastq_fp_list[file_id].write(line1)
                out_fastq_fp_list[file_id].write(line2)
                out_fastq_fp_list[file_id].write(line3)
                out_fastq_fp_list[file_id].write(line4)

            out_readname_set.add(readname)

        in_fastq_fp.close()

    for out_fastq_fp in out_fastq_fp_list:
        out_fastq_fp.close()

    return out_fastq_file_list

def extract_fastq_tail_seq(in_fastq_file, read_tail_length, left_tail_fastq_file, right_tail_fastq_file):

    in_fastq_fp = gzopen(in_fastq_file)
    fq_left_tail_fp  = open(left_tail_fastq_file, 'w')
    fq_right_tail_fp = open(right_tail_fastq_file, 'w')

    while 1:
        line1 = in_fastq_fp.readline()
        line2 = in_fastq_fp.readline()
        line3 = in_fastq_fp.readline()
        line4 = in_fastq_fp.readline()

        if not line1: break
        if not line2: break
        if not line3: break
        if not line4: break

        if line1[0] != '@' or len(line2) != len(line4) or line3.strip() != '+':
            eprint('ERROR! bad fastq record: ')
            eprint(line1 + line2 + line3 + line4)
            sys.exit()
        
        read_seq  = line2.strip()
        read_qual = line4.strip()

        left_tail_seq  = read_seq[0:read_tail_length]
        left_tail_qual = read_qual[0:read_tail_length]

        right_tail_seq  = rev_comp(read_seq[-read_tail_length:])
        right_tail_qual = ''.join(reversed(read_qual[-read_tail_length:]))

        fq_left_tail_fp.write(line1)
        fq_left_tail_fp.write(left_tail_seq + '\n')
        fq_left_tail_fp.write(line3)
        fq_left_tail_fp.write(left_tail_qual + '\n')

        fq_right_tail_fp.write(line1)
        fq_right_tail_fp.write(right_tail_seq + '\n')
        fq_right_tail_fp.write(line3)
        fq_right_tail_fp.write(right_tail_qual + '\n')

    in_fastq_fp.close()
    fq_left_tail_fp.close()
    fq_right_tail_fp.close()

    return


def rev_comp(seq):

    complement  = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'a': 't', 'c': 'g', 'g': 'c', 't': 'a'}
    rev_seq = ''.join(reversed(seq))

    rev_comp_seq = ''
    for b in rev_seq:
        rev_comp_seq += complement[b]

    return rev_comp_seq



def format_parameters(input_para):

    input_para = input_para.strip('`')
    input_para = input_para.strip('\'')
    input_para = input_para.strip('\"')
    input_para = input_para.lower()

    return input_para

def compute_overlap_len(start1, end1, start2, end2):
    max_start = max(start1, start2)
    min_end   = min(end1, end2)

    overlap_len = min_end - max_start
    return max(0, overlap_len)



### SAM/BAM/Alignment ###


def analysis_cigar_string (cigar):
    
    opr_set   = set(['=', 'X', 'I', 'D', 'N', 'S', 'H', 'P', 'M'])
    digit_set = set(['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'])

    cigar_opr_list = list()
    cigar_opr_len_list = list()
    
    length_str = ''
    for i in range(0, len(cigar)):
        c = cigar[i]
        if c in digit_set:
            length_str += c
        elif c in opr_set:
            cigar_opr_list.append(c)
            cigar_opr_len_list.append(int(length_str))
            length_str = ''
        else:
            eprint('ERROR: unknown CIGAR operation: %s' % c)
            sys.exit()
    
    return cigar_opr_list, cigar_opr_len_list



def calculate_repeat_size_from_exact_match(cigar, tstart, ref_repeat_start_pos, repeat_unit_size):
    repeat_size = 0
    cigar_opr_list, cigar_opr_len_list = analysis_cigar_string(cigar)
    current_ref_pos = tstart

    for i in range(0, len(cigar_opr_list)):
        cigar_opr = cigar_opr_list[i]
        cigar_opr_len = cigar_opr_len_list[i]
        if cigar_opr == '=':  # match
            if current_ref_pos >= ref_repeat_start_pos:
                repeat_size += int(cigar_opr_len/repeat_unit_size)
            elif current_ref_pos + cigar_opr_len >= ref_repeat_start_pos:
                overlap_len = current_ref_pos + cigar_opr_len - ref_repeat_start_pos
                if overlap_len > 0: 
                    repeat_size += int(overlap_len/repeat_unit_size)
        
            current_ref_pos += cigar_opr_len
        elif cigar_opr == 'X': # mismatch
            current_ref_pos += cigar_opr_len
        elif cigar_opr == 'I': # insertion
            pass
        elif cigar_opr == 'D': # deletion
            current_ref_pos   += cigar_opr_len
        else:
            eprint('ERROR! unsupported CIGAR operation: %s' % cigar_opr)
            sys.exit()

    return repeat_size


def target_region_alignment_stats_from_cigar(cigar, tstart, tend, ref_region_start, ref_region_end):

    # tstart is the aligned start position in the paf file
    # ref_region_start and ref_region_end are the target region in the reference

    if len(cigar) == 0:
        eprint('ERROR! cigar string is empty!')
        sys.exit()
    
    matching_score = 2
    mismatching_penalty = -4
    gap_open_penalty = -4
    gap_ext_penalty = -2

    num_match = 0
    num_mismatch = 0
    num_ins = 0
    num_del = 0
    align_score = 0
    cigar_opr_list, cigar_opr_len_list = analysis_cigar_string(cigar)
    if len(cigar_opr_list) != len(cigar_opr_len_list):
        eprint('ERROR: len(cigar_opr_list) != len(cigar_opr_len_list) This could be a bug.')
        sys.exit()
        
    current_ref_pos = tstart
  
    for i in range(0, len(cigar_opr_list)):
        cigar_opr = cigar_opr_list[i]
        cigar_opr_len = cigar_opr_len_list[i]
        if cigar_opr == '=':  # match
            overlap_len = compute_overlap_len(current_ref_pos, current_ref_pos+cigar_opr_len, ref_region_start, ref_region_end)
            if overlap_len > 0: 
                num_match += overlap_len
                align_score += overlap_len * matching_score
            current_ref_pos += cigar_opr_len
        elif cigar_opr == 'X': # mismatch
            overlap_len = compute_overlap_len(current_ref_pos, current_ref_pos+cigar_opr_len, ref_region_start, ref_region_end)
            if overlap_len > 0: 
                num_mismatch += overlap_len
                align_score += overlap_len * mismatching_penalty
            current_ref_pos += cigar_opr_len
        elif cigar_opr == 'I': # insertion
            if current_ref_pos > ref_region_start and current_ref_pos < ref_region_end - 1:
                num_ins += cigar_opr_len
                align_score += gap_open_penalty + (cigar_opr_len-1) * gap_ext_penalty
        elif cigar_opr == 'D': # deletion
            overlap_len = compute_overlap_len(current_ref_pos, current_ref_pos+cigar_opr_len, ref_region_start, ref_region_end)
            if overlap_len > 0:
                num_del += overlap_len
                align_score += gap_open_penalty + (overlap_len-1) * gap_ext_penalty
            current_ref_pos += cigar_opr_len
        elif cigar_opr == 'S':
            continue
        else:
            eprint('ERROR: unsupported cigar operation: %s' % cigar_opr)
            sys.exit()
        if current_ref_pos > ref_region_end: break

    if tend < ref_region_end:
        num_mismatch += ref_region_end - tend
    
    if tstart > ref_region_start:
        num_mismatch += tstart - ref_region_start

    read_error = ReadError(num_match, num_mismatch, num_ins, num_del, align_score)
    return read_error

def get_preset_for_minimap2(data_type):
    if data_type == 'ont':
        preset = ' -x map-ont '
    elif data_type == 'ont_sup':
        preset = ' -x map-ont '
    elif data_type == 'ont_q20':
        preset = ' -x map-ont '
    elif data_type == 'clr':
        preset = ' -x map-ont '
    elif data_type == 'hifi':
        preset = ' -x map-ont '
    else:
        eprint(f'ERROR: Unknown data type: {data_type}\n')
        sys.exit(1)
    
    return preset

minimap2_mid_para   = ' -k 5 -w 3 -n 1 -m 10 -s 40 '
minimap2_short_para = ' -k 3 -w 2 -n 1 -m 10 -s 40 '
### system ###

def switch_two_numbers(a, b):
    return b, a

def switch_two_objects(a, b):
    return b, a

def run_system_cmd(cmd):
    eprint(f'NOTICE: Running command: {cmd}')  
    ret = os.system(cmd)
    if ret != 0: 
        eprint('ERROR: Failed to run command: %s' % cmd)
        eprint('Return value is: %d' % ret)
        sys.exit()
    return

def find_executable_path(name):
    path = distutils.spawn.find_executable(name)
    return path

def rm(path):
    if os.path.exists(path):
        try:
            os.remove(path)
        except:
            eprint(f'WARNING: Failed to remove file: {path}')
    return
